#include <stdio.h>                  // printf()
#include "..\types\types.h"
#include "..\dpmi32\phys.h"
#include "..\intel\timestmp.h"


#define TEST_MASK       0x87654321
#define INCREMENT       4



int main( void )
{
    // test code for physical memory addressing
    //
    PentiumClock start, end, delta;
    double bandwidth;
    Bit32 *io_buffer;
    Bit32 mem_handle;
    Bit32 count, found;

    // map 57M of physical memory from 610000 to top of memory into our
    // process's address space so that it's addressable via DS
    //
    io_buffer = (Bit32 *) map_physical( PHYS_RAM_START, PHYS_RAM_LENGTH, &mem_handle );

    start.set();

    // touch every DWORD and give it a unique value
    //
    for( count=0; count<PHYS_RAM_LENGTH/sizeof(Bit32); count++ )
        io_buffer[ count ] = TEST_MASK + count;

    // check every DWORD on a separate pass
    // to make sure it held the values
    //
    for( count=0; count<PHYS_RAM_LENGTH/sizeof(Bit32); count++ )
    {
        found = io_buffer[ count ];
        if( found != TEST_MASK + count )
            break;
    }

    end.set();

    delta = end - start;
//  printf( "start   time was 0x%8.8lX%8.8lX\n", start.high, start.low );
//  printf( "end     time was 0x%8.8lX%8.8lX\n",   end.high,   end.low );
    printf( "elapsed time was 0x%8.8lX%8.8lX = %g CPU clocks = %g seconds\n",
        delta.high, delta.low, (double)delta, ((double)delta)/CLOCK_RATE );

    bandwidth = (double) count * 2 * CLOCK_RATE / (double) delta;
    printf( "memory bandwidth was %g words per second\n", bandwidth );

    if( count < PHYS_RAM_LENGTH/sizeof(Bit32) )
    {
        Bit32 logical = (Bit32) (io_buffer + count);
        Bit32 linear = get_base( _DS ) + logical;
        Bit32 physical = PHYS_RAM_START + count*sizeof(Bit32);

        printf( "memory compare failure at logical 0x%4.4X:%8.8lX"
            " = linear 0x%8.8lX = physical 0x%8.8lX\n"
            "value 0x%8.8lX should have been 0x%8.8lX\n",
            _DS, logical, linear, physical,
            found, TEST_MASK-count
        );

    }
    else
        printf( "successfully looked at %ld memory locations (%ld bytes) in strides of %d bytes\n", count, count*sizeof(Bit32), INCREMENT );

    // cleanup for physical memory access
    //
    unmap_physical( PHYS_RAM_LENGTH, mem_handle );

    return 0;
}
