#ifndef _PCA200E_H
#define _PCA200E_H



#include "..\fore\types.h"              // u_int, Vpvc, Aal_type, Haddr
#include "..\fore\fatal.h"              // heartbeat fatal error codes
#include "..\fore\shared_m.h"           // offsets to card data structures
#include "..\fore\ncomm.h"              // soft UART structure, access macros
#include "..\fore\board.h"              // mon960 communication area
#include "..\fore\aali.h"               // commands and their data structures
#include "..\integrat\constant.h"       // whether cell payload got copied to host and for which cells



// definitions of terms:
//
//      i960    Intel 80960CA RISC 50 MHz microprocessor
//      ESP     enhanced SAR processor (ASIC between PHY and i960)
//      PBI     PCI bus interface (ASIC between i960 and PCI)
//      PCI     peripheral component interconnect (local bus)
//      PHY     SONET payload cell extraction/insertion device
//      SAR     segmentation and reassembly
//



// only defined for reading
//
// see HostControl_write if you are writing
//
struct HostControl_read
{
    Bit8 board_reset        : 1;        // cleared 63 bus cycles after PCI-bus-caused reset
    Bit8 lock_hold          : 1;        // you did this to me!
    Bit8 i960_fail          : 1;        // i960 failed its internal self test
    Bit8 hold_ack           : 1;        // current state of i960 HOLDA line
    Bit8 out_fifo_full      : 1;        // i960 -> PBI FIFO is totally full (so i960 held)
    Bit8 esp_hold           : 1;        // ESP is holding i960
    Bit8 in_fifo_almost_full : 1;       // PBI -> i960 FIFO is almost full (so maybe deadlock)
    Bit8 test_mode          : 1;        // PBI is currently in its loopback test mode
};



// only defined for writing
//
// see HostControl_read if you are reading
//
struct HostControl_write
{
    Bit8 board_reset        : 1;        // set then clear to cause reset of all chips on board
    Bit8 lock_hold          : 1;        // 1 prevents i960 from executing beyond this access, 0 allows i960 to run
    Bit8 set_i960_int_2     : 1;        // 1 causes slave interrupt B to i960, 0 has no effect
    Bit8 set_i960_int_1     : 1;        // 1 causes slave interrupt A to i960, 0 has no effect
    Bit8 clear_host_int     : 1;        // 1 clears PCI->host interrupt, 0 has no effect
    Bit8                    : 3;        // unused (?ignored) during write
};



// is both readable and writable by host
//
struct HostInterruptMask
{
    Bit8 is_masked          : 1;
    Bit8                    : 7;
};



// no effect if host writes (I think)
//
struct PciSpecific_read
{
    Bit8 is_interrupting    : 1;
    Bit8                    : 7;
};



struct SharedMemory
{
    Bit8 unknown1[ MON960COM_OFFSET ];
    Mon960com mon960;

    // should really use start_offset field in FirmwareHeader
    // but that isn't known until runtime, and this is just as
    // safe
    //
    Bit8 unknown2[ COMMON_ORIGIN - MON960COM_OFFSET - sizeof(Mon960com) ];
    Cp_queues queues;
};



// definitions of a FORE PCA200E PCI SAR
// as it is seen in host memory
//
struct pca200e_memmap
{
    union
    {
        SharedMemory shared;
        Bit8  ram8[ 0x100000 ];
        Bit16 ram16[ 1 ];
        Bit32 ram32[ 1 ];
    };

    union
    {
        HostControl_read hcr_r;
        HostControl_write hcr_w;
        Bit32 dword_00h;
    };

    union
    {
        HostInterruptMask himr;
        Bit32 dword_01h;
    };

    union
    {
        PciSpecific_read psr;
        Bit32 dword_02h;
    };
};



#define FIRMWARE_PATH           "\\oc3mon\\"
#if CELLS_PER_BUFFER > 0
    #if HOST_MUST_REASSEMBLE
        #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "pay_all.bin"
    #else
        #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "pca200e.bin"
    #endif
#else
    #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "only_atm.bin"
#endif



// header in front of a firmware binary image file
//
struct FirmwareHeader
{
    Bit32 id;
    Bit32 version_id;
    Bit32 start_offset;         // also points at Cp_queues
    Bit32 entry_address;
};



#endif              // not included yet
