/* 1325, Tue 20 Aug 96

   METER_PC.C:  The AU accounting meter mainline;
		   based on CMU's snmpd,
		   PCIP's packet monitoring
		   and Waterloo's TCP/IP.

   Copyright (C) 1992-1996 by Nevil Brownlee,
   ITSS Technology Development,  The University of Auckland */

/*
 * snmpd.c - send snmp GET requests to a network entity.
 *
 */
/***********************************************************
	Copyright 1988, 1989 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#define Q_DEBUG  0

extern char version_descr[];  /* In met_vars.c */
extern unsigned int SampleRate;
extern char *communities[];  /* In snmp/snmpagnt.c */

char interface_name[6];  /* Device name for metered interface */

#include <dos.h>
#include <conio.h>
#include <alloc.h>
#include <ctype.h>
#include <errno.h>

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include "tcp.h"
#include "pktdrv.h"

#include "ausnmp.h"
#include "snmp.h"
#include "asn1.h"
#include "snmpimpl.h"
#include "snmpapi.h"

#define PKTSNAP
#include "pktsnap.h"
#include "flowhash.h"
#include "met_vars.h"

extern int errno;
int snmp_dump_packet = 0;

udp_Socket udpsock;
static udp_Socket *s;
int status;  /* For socket */
unsigned long gethostid(void);  /* In src\pcbsd.c */
void init_snmp(void);  /* In snmp\snmpapi.c */

void zero_pkt_stats(void);  /* meter_pc function prototypes */
void receive(void);
void meter_snmp_read(int length);

unsigned int sample_count = 1;
int n_interfaces, IP_int;

void accept_pkt(int len, unsigned char far *buff, int if_nbr)
{
   unsigned int ether_type, lsap;
   unsigned char far *p;
   struct pkt huge *pp;
   int pkt_ix;

   if (sample_count != 1)  /* Process only 1 pkt per SampleRate */
      --sample_count;
   else {
      sample_count = SampleRate;
      ++en_table[if_nbr].npackets;
      en_table[if_nbr].nbytes += len;
      asm pushf;  /* Save flags (especially int flag) */
      asm cli;    /* Disable interrupts */
	 if ((pkt_ix = free_stack) != 0) {
	    pp = &pkts[pkt_ix];
	    free_stack = pp->next;
	    }
      asm popf;   /* Restore flags */
      if (pkt_ix == 0) {  /* No pkts free, ignore the data */
	 ++lostpackets;
	 ++en_table[if_nbr].lostpackets;
	 }
      else {
	 if (buff[12] <= 5) {  /* 802.3 (5FF hex = 1535 decimal) */
	    lsap = buff[14] << 8 | buff[15];
	    if (lsap == 0xAAAA) {  /* SNAP packet */
	       ether_type = buff[20] << 8 | buff[21];
	       p = &buff[22];
	       }
	    else {
	       ether_type = 0;
	       p = &buff[14];
	       }
	    }
	 else {   /* Blue book */
	    lsap = 0;
	    ether_type = buff[12] << 8 | buff[13];
	    p = &buff[14];
	    }
	 pp->p_len = len;
	 pp->Interface = if_nbr+1;  /* 1-org */
	 pp->arrival_time = uptime();
	 handle_pkt(pp, ether_type,lsap, buff,p);
	 pp->next = 0;
	 asm pushf;  /* Save flags (especially int flag) */
	 asm cli;    /* Disable interrupts */
	    if (proc_n_pkt == 0)  /* Empty queue */
	       proc_head = pkt_ix;
	    else pkts[proc_tail].next = pkt_ix;
	    proc_tail = pkt_ix;
	    ++proc_n_pkt;
	 asm popf;   /* Restore flags */
	 }
      }
   }

#define SNAPSIZE  54  /* Matches meter_ux.c */

int our_packet(unsigned char far *buff)
{
   if (buff[0] == 0xFF) {  /* Broadcast */
      if (buff[13] == 0x06 && buff[12] == 0x08)  /* ARP */
	 return 1;
      }
   else if (buff[5] == eth_addr[5] && buff[4] == eth_addr[4] &&
	 buff[3] == eth_addr[3] && buff[2] == eth_addr[2] &&
	 buff[1] == eth_addr[1] && buff[0] == eth_addr[0]) {
      if (buff[12] == 0x08 &&
	    (buff[13] == 0x00 || buff[13] == 0x06 )) {  /* IP or ARP */
	 return 1;
	 }
      }
   return 0;  /* Only want our IP & ARP pkts */
   }

unsigned char far *pkt_rcv_call1(  /* Nevil, 15 Jul 96 */
   int len, int lah_len, unsigned char far *buff, int if_nbr)
{
   int if_type = en_table[if_nbr].type;
   if (if_type & EN_HP) {
      if (if_type & EN_MT)
	 accept_pkt(len, buff,if_nbr);  /* Pass attrib info to meter buffer */
      if (if_type & EN_IP) {
	 if (our_packet(buff)) return get_pktbuf(len);
	 }
      return (unsigned char far *)NULL;  /* Don't bother with call2 */
      }
   else  /* Can't see packet contents until call2 */
      return get_pktbuf(len);
   }

void pkt_rcv_call2(int len, unsigned char far *buff, int if_nbr)
{
   int if_type = en_table[if_nbr].type;
   if (if_type & EN_HP)
      buff[-2] = 1;  /* IP pkt ready for tcp/ip */
   else {
      if (if_type & EN_MT) {
	 accept_pkt(len, buff,if_nbr);  /* Pass attribute info to meter buffer */
	 if (if_type & EN_IP)
	    buff[-2] = our_packet(buff) ? 1 : 0;
	 }
      else if (if_type & EN_IP) {
	 buff[-2] = 1;  /* Unicast/broadcast mode */
	 }
      else buff[-2] = 0;  /* Not our packet, free the buffer */
      }
   }

void au_send_mon(int length, unsigned char far *buffer)
{  /* Invoked by pkt_send in src/pcpkt.c (if monitoring on IP interface) */
   accept_pkt(length, buffer,IP_int);  /* $$$ */
   }

void main(int argc, char *argv[])
{
   int arg, a, c;
   char cbuf[20], *ap;
   struct en_info *enp;

   hwt_init();  /* Initialise hwtimer routines */
   sclear();
   scpos(0,0);  printf(version_descr);
   scpos(0,1);

   en_table[0].type = EN_IP+EN_MT;  /* $$$ */
   maxpkt = 1024;
   n_interfaces = 0;  IP_int = -1;
   display_enabled = kb_enabled = 1;  /* Enabled by default */
   chart_interface = 0;
   mxflowsp1 = DFMXFLOWS+1;
   /* Default (CMU) communities are:  0 = "public", 1 = "proxy",
	 2 = "private", 3 = "regional", 4 = "core"
      We only allow "public" and "private" by default */
   communities[1] = communities[3] = communities[4] = "";
   for (c = 0, arg = 1; arg < argc; ++arg) {
      if (argv[arg][0] == '-') {
	 ap = argv[arg]+2;
	 switch (a = argv[arg][1]) {
	 case 'd':
	    snmp_dump_packet++;
	    break;
	 case 'f':
	    if (*ap == NULL) ap = argv[++arg];
	    mxflowsp1 = atoi(ap)+1;
	    break;
	 case 'I':  /* IP only - no metering */
	    IP_int = n_interfaces;
	 case 'i':  /* Just plain metering */
	 case 'h':  /* Metering with high-performance driver */
	    if (*ap == NULL) ap = argv[++arg];
	    en_table[n_interfaces].int_nbr = atoi(ap);
	    if (a != 'I') en_table[n_interfaces].type =
	       a == 'h' ? EN_MT|EN_HP : EN_MT;
	    ++n_interfaces;
	    break;
	 case 'k':
	    kb_enabled = 0;  /* -k to disable keyboard */
	    break;
	 case 'n':
	    if (*ap == NULL) ap = argv[++arg];
	    SampleRate = atoi(ap);
	    if (SampleRate < 1) SampleRate = 1;
	    break;
	 case 'p':
	    if (*ap == NULL) ap = argv[++arg];
	    maxpkt = atoi(ap);
	    break;
	 case 'r':
	    if (*ap == NULL) ap = argv[++arg];
	    communities[0] = ap;  /* -r to set read community */
	    break;
	 case 's':
	    display_enabled = 0;  /* -s to disable screen */
	    break;
	 case 'w':
	    if (*ap == NULL) ap = argv[++arg];
	    communities[2] = ap;  /* -w to set write community */
	    break;
	 default:
	    printf("Invalid option: -%c\n", argv[arg][1]);
	    hwt_fin();   /* Restore BIOS timer routines */
	    exit(0);
	    }
	 }
      }
   if (n_interfaces == 0) {
      en_table[0].type= EN_MT;  n_interfaces =  1;
      }
   if (IP_int == -1) IP_int = 0;
   en_table[IP_int].type |= EN_IP;

   pkts = (struct pkt far *)farmalloc((maxpkt+1) * sizeof(struct pkt));
   for (c = 1; c <= maxpkt; ++c)  /* pkt[0] used for dummy packets */
      pkts[c].next = c-1;
   free_stack = maxpkt;
   proc_n_pkt = proc_head = proc_tail = 0;

   sock_init();
   printf("Running on %s",inet_ntoa(cbuf,gethostid()));
   sprintf(interface_name, "et%02X",en_table[0].int_nbr);

   s = &udpsock;  /* myport        hisadr hisport */
   if (!udp_open( s, SNMP_PORT,  0, 0,        0 )) {
      puts("udp_open failed!\n");
      hwt_fin();   /* Restore BIOS timer routines */
      exit(1);
      }

   init_snmp();
   init_monitor();
   AdjAddrType = 1;  /* Ethernet */
   receive();
   }

void show_meter_time()
{
   char msg[60];
   sprintf(msg,"%lu seconds since %02d%02d:%02d",
      elapsed_sec, s_tod_h,s_tod_m,s_tod_s);
   display_msg(0,msg);
   }

void set_chart_interface(int int_nbr)
{
   chart_interface = int_nbr;
   spackets = en_table[chart_interface].npackets;
   sbytes = en_table[chart_interface].nbytes;
   printf("Charting interface %d", int_nbr+1);
   }

unsigned int mxmonpkts = 400;  /* Max nbr of packets per pkt_monitor call */
unsigned int bkgi;  /* Seconds before next run of backgound process */
#define BKG_INTERVAL  30  /* 30 seconds */

void zero_pkt_stats()
{
   int j;
   struct en_info *eip;
   for (j = 0; j != n_interfaces; ++j) {
      eip = &en_table[j];
      eip->npackets = eip->nbytes = eip->lostpackets = 0L;
      }
   badpackets = nobufpackets = lostpackets  =
      stats_time = t_backlog = spackets = sbytes =
      kilodummypackets = dummypacketrate = 0L;
   mindummyrate = 100000000L;
   pkt_backlog = max_pkt_backlog = max_pkt_rate =
      dummypackets = mdpacketrate = 0;
   clear_pkt_stats = 0;
   }

void receive()
{
   unsigned char startup, ch, gci, p;
   unsigned int k, pkt_ix, n, x, sumx, minx, maxx, samples, ax;
   char l_tod_s;  /* Last tod_s displayed on chart */
   struct pkt huge *pp;
   struct en_info *eip;
   unsigned long pd,bd;
   long tot = set_timeout(1);
   char huge *cp;
   unsigned char half_scale = 0;

   startup = 1;
   for (;;) {
      sock_tick(s, &status);
      if ((n = sock_dataready(s)) != 0) {
	 meter_snmp_read(n);
	 sock_close(s);
	 if (!udp_open( s, SNMP_PORT,  0, 0,   0 )) {
            puts("udp_open failed!\n");
	    hwt_fin();   /* Restore BIOS timer routines */
	    exit(1);
            }
	 }

      if (proc_n_pkt == 0) {  /* Build a dummy packet */
	 cp = (char far *)&pkts[0];
	 for (k = 0; k != sizeof(struct pkt); ++k) cp[k] = 0;
	 pkts[0].PeerAddrType = AT_DUMMY;
	 pkt_monitor(&pkts[0]);
	 }
      else {
	 if (proc_n_pkt > pkt_backlog) pkt_backlog = proc_n_pkt;
	 for (k = mxmonpkts; proc_n_pkt != 0 && k != 0; --k) {
	    asm pushf;  /* Save flags (especially int flag) */
	    asm cli;    /* Disable interrupts */
	       pp = &pkts[pkt_ix = proc_head];
	       proc_head = pp->next;
	       if (--proc_n_pkt == 0) proc_tail = 0;
	    asm popf;   /* Restore flags */
            pkt_monitor(pp);
	    asm pushf;  /* Save flags (especially int flag) */
	    asm cli;    /* Disable interrupts */
	       pp->next = free_stack;
	       free_stack = pkt_ix;
	    asm popf;   /* Restore flags */
	    }
	 }

      if (chk_timeout(tot)) {
	 tot = set_timeout(1);  /* 1 second */
	 check_uptime();
	 ++elapsed_sec;  set_tod();
	 eip = &en_table[chart_interface];
	 if (startup) {  /* Running now, initialise counters etc. */
	    clear_pkt_stats = 1;
	    minx = 100;  sumx = samples = maxx = 0;
	    spackets = eip->npackets;  sbytes= eip->nbytes;
	    l_tod_s = -1;
	    gci = gc_interval;  /* Garbage collect interval */
	    bkgi = BKG_INTERVAL;  /* Background interval */
	    gc_f = (mxflowsp1-1)/100;  /* Flow indexes to check */
	    if (gc_f < 4) gc_f = 4;
	    startup = 0;
	    }
	 else {
            if (--gci == 0) {
	       garbage_collect(1);  /* Routine incremental collection */
               gci = gc_interval;
               }
	    if (--bkgi == 0) {  /* Check % of flows active */
	       p = (unsigned long)active_flows()*100/(mxflowsp1-1);
	       if (display_enabled) {
		  scpos(12,3);  printf("a=%3u",p);
		  }
	       if (p > FloodMark &&
		     FloodMark > 0 && FloodMark < 100) {
		  if (CurrentRuleSet != 1) {
		     close_rule_set();
		     open_rule_set(1,1);
		     display_msg(1,"Switched to default rules");
		     }
		  }
	       else if (p > HighWaterMark &&
		     HighWaterMark > 0 && HighWaterMark < 100 &&
		     EmergencyRuleSet != 0 &&
		     CurrentRuleSet != EmergencyRuleSet) {
		  close_rule_set();
		  open_rule_set(1,EmergencyRuleSet);
		  display_msg(1,"Switched to emergency rules");
		  }
	       else if (p > HighWaterMark*2/3)
		  more_garbage();
	       bkgi = BKG_INTERVAL;
	       }
	    npackets = eip->npackets;
	    pd = (eip->npackets-spackets)*SampleRate;
	    bd = (eip->nbytes-sbytes)*SampleRate;
	    if (display_enabled) {
   	       scpos(0,3);
	       printf("q=%4d",pkt_backlog);
	       scpos(23,3);
	       printf("%02d%02d:%02d", tod_h,tod_m,tod_s);
	       scpos(0,5);
	       x = util_pc(pd,bd);  /* % utilisation */
	       printf("p=%6lu, b=%10lu, u=%4d", pd,bd, x);
	       sumx += x;  ++samples;
	       if (x < minx) minx = x;
	       if (x > maxx) maxx = x;
	       if (tod_s != l_tod_s && tod_s%10 == 0) {
	          if (samples) {  /* 30% in col 79 */
                     ax = (sumx+(samples>>1))/samples;
                     if (half_scale) 
                        chart(41,0, 79,24, minx>>1,ax>>1,maxx>>1);
		     else chart(41,0, 79,24, minx,ax,maxx);
		     l_tod_s = tod_s;
		     }
	          minx = 100;  sumx = samples = maxx = 0;
		  }
               }
	    }
	 if (clear_pkt_stats) zero_pkt_stats();
	 else {
            ++stats_time;
            if (pd > max_pkt_rate) max_pkt_rate = pd;
            t_backlog += (unsigned long)pkt_backlog;
	    if (pkt_backlog > max_pkt_backlog)
	       max_pkt_backlog = pkt_backlog;
	    if (dummypacketrate < mindummyrate) {
	       mindummyrate = dummypacketrate;  mdpacketrate = pd;
	       }
	    }

	 pkt_backlog = 0;  dummypacketrate = 0;
	 spackets = eip->npackets;  sbytes = eip->nbytes;

	 if (kb_enabled && kbhit()) {
	    if ((ch = getch()) == 27) {  /* ESC */
	       scpos(0,24);
	       printf("\nShutting down\n");
	       hwt_fin();   /* Restore BIOS timer routines */
	       exit(0);
	       }
	    w_clear(0,24, 40,24);  scpos(0,24);
	    switch (k = tolower(ch)) {
	    case 'b':
	       printf("bad=%lu, nobuf=%lu, lost=%lu",
		  badpackets,nobufpackets,lostpackets);
	       break;
	    case 'h':
	       half_scale = !half_scale;
	       printf("Chart scale 0..%d \%",
		  half_scale ? 60 : 30);
	       break;
	    case 'i':
	       printf("Ethernet %d, type %d, sw int %02X",
		  chart_interface+1,
		  en_table[chart_interface].type,
		  en_table[chart_interface].int_nbr);
	       break;
	    case 'm':
#if defined(__LARGE__) | defined(__HUGE__) | defined(__COMPACT__)
	       /* Large models don't bother with a near heap */
	       printf("%lu bytes free", farcoreleft());
#else
	       printf("%u near, %lu far bytes free",
		  coreleft(),farcoreleft());
#endif
	       break;
	    case 'v':
	       printf(version_descr);
	       break;
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	       set_chart_interface(k - '1');
	       break;
	    default:
	       handle_kb(ch);
	       break;
	       }
	    }
         }
      }

sock_err:
   switch (status) {
   case 1 : /* foreign host closed */
      hwt_fin();   /* Restore BIOS timer routines */
      exit(0);
   case -1: /* timeout */
      printf("\nConnection timed out!");
      hwt_fin();   /* Restore BIOS timer routines */
      exit(1);
   default: printf("Aborting");
      exit(1);
      }
}

unsigned long snmp_peer_addr;

void meter_snmp_read(int length)
{
   int out_length;
   u_char packet[1500], outpacket[1500];
   int count;
   struct sockaddr from;
   int fromlen = sizeof(struct sockaddr);
   char snmp_peer_name[30];  /* name of host which sent the snmp request */
   getpeername(s, &from,&fromlen);
   snmp_peer_addr = from.s_ip;

   sock_read(s, packet,length);

   if (snmp_dump_packet) {
      inet_ntoa(snmp_peer_name,from.s_ip);
      printf("received %d bytes from %s:\n", length, snmp_peer_name);
      for (count = 0; count < length; count++) {
	 printf("%02X ", packet[count]);
	 if ((count % 16) == 15) printf("\n");
	 }
      printf("\n\n");
      }
   out_length = 1500;
   if (snmp_agent_parse(packet, length, outpacket, &out_length, from.s_ip)) {
      if (snmp_dump_packet) {
	 printf("sent %d bytes to %s:\n", out_length, snmp_peer_name);
	 for (count = 0; count < out_length; count++) {
	    printf("%02X ", outpacket[count]);
	    if ((count % 16) == 15) printf("\n");
	    }
         printf("\n\n");
	 }
      sock_write(s, (char *)outpacket, out_length);
      }
   }
