#include <stdio.h>              // printf()
#include <stdlib.h>             // exit()
#include "pci.h"



typedef Bit32 (*reader_t)( Bit16 handle, Bit16 offset );



int main( int argc, char *argv[] )
{
    Bit16 handle, address, start, length, line;
    Bit32 mask;
    Bit32 value;
    int width;
    const char *program_name = argv[0];
    reader_t reader;

    // parse command-line arguments
    //
    if( argc != 5 )
    {
        fprintf( stderr, "usage: %s card-handle config-offset bit-width config-length\n", program_name );
        return 1;
    }
    handle = strtoul( argv[1], NULL, 0 );
    start  = strtoul( argv[2], NULL, 0 );
    length = strtoul( argv[4], NULL, 0 );
    width  = atoi( argv[3] );

    switch( width )
    {
        default:
            fprintf( stderr, "Sorry, you can only use widths of 8, 16, or 32 bits.\n" );
            return 3;

        case 8:
            reader = (reader_t) pci_read_config_8;
            mask = 0xFF;
            break;

        case 16:
            reader = (reader_t) pci_read_config_16;
            mask = 0xFFFF;
            break;

        case 32:
            reader = pci_read_config_32;
            mask = 0xFFFFFFFF;
            break;
    }

    // check whether PCI BIOS is present
    //
    if( !pci_is_bios_present() )
    {
        printf( "Sorry, PCI BIOS is not present.\n" );
        return 2;
    }

    printf( "dumping %d-bit PCI config values starting at offset 0x%.4X, length 0x%.4X\n", width, start, length );

    for( address=start; length>15; address+=16,length-=16 )
    {
        printf( "%.4X  ", address );

        for( line=0; line<16; line+=width/8 )
        {
            value = mask & reader( handle, address + line );
            printf( " %.*X", width/4, value );
        }

        printf( "\n" );
    }

    return 0;
}
