#ifndef _TYPES_H
#define _TYPES_H



#include <limits.h>                     // ULONG_MAX UINT_MAX



#define TRUE  1
#define FALSE 0



#if defined(SUNOS)
#   define HOST_IS_BIG_ENDIAN TRUE
#else
#   define HOST_IS_BIG_ENDIAN FALSE
#endif


typedef unsigned char Bit8;
typedef unsigned short Bit16;

#if __alpha
    typedef unsigned Bit32;
    typedef unsigned long Bit64;
#   define BIT64_AVAILABLE TRUE
#   define BIT32_MAX UINT_MAX

#else
    typedef unsigned long Bit32;
#   define BIT64_AVAILABLE FALSE
#   define BIT32_MAX ULONG_MAX

#   ifdef __cplusplus

#       define DWORD_SIZE    4294967296.0      // 4G

        // 64-bit integer, for use when your compiler does
        // not support such a thing intrinsically
        //
        class Bit64
        {

        public:
#if HOST_IS_BIG_ENDIAN
            Bit32 high;
            Bit32 low;
#else
            Bit32 low;
            Bit32 high;
#endif

            // no destructor needed

            // construct w/o args
            //
            Bit64()
            {
            }

#if 1
            // construct from unsigned 32-bit integer
            //
            Bit64( Bit32 second ) :
                low( second ),
                high( 0 )
            {
            }
#else
            // construct from unsigned 32-bit integer
            //
            Bit64( Bit32 second )
            {
                *this = second;
            }
#endif
            // construct from double-precision floating
            // point using that assignment operator
            //
            Bit64( double second )
            {
                *this = second;
            }

            // assignment from double-precision floating point operator
            //
            Bit64 &operator =( double second )
            {
                high = second / DWORD_SIZE;
                low  = second - (high * DWORD_SIZE);
                return *this;
            }

            // assignment from unsigned 32-bit integer
            //
            Bit64 &operator =( Bit32 second )
            {
                high = 0;
                low  = second;
                return *this;
            }

#if 0
            // type conversion to single-precision floating point
            //
            operator float() const
            {
                return high * (float)DWORD_SIZE + low;
            }
#endif

            // type conversion to double-precision floating point
            //
            operator double() const
            {
                return high * (double)DWORD_SIZE + (double)low;
            }

            // subtraction operator overload
            //
            Bit64 operator -( const Bit64 &second ) const
            {
                Bit64 result;

                result.low = low - second.low;
                result.high = high - second.high;

                // if there was a carry...
                //
                if( result.low > low )

                    // believe it or not, the low word
                    // is already correct!!!
                    //

                    // fix the high dword
                    //
                    result.high--;

                return result;
            }

            // subtraction with assignment operator overload
            //
            Bit64 &operator -=( const Bit64 &second )
            {
                Bit32 low2 = low - second.low;
                high -= second.high;

                // if there was a carry...
                //
                if( low2 > low )

                    // believe it or not, the low word
                    // is already correct!!!
                    //

                    // fix the high dword
                    //
                    high--;

                low = low2;
                return *this;
            }

            // addition operator overload
            //
            Bit64 operator +( const Bit64 &second ) const
            {
                Bit64 result;

                result.low = low + second.low;
                result.high = high + second.high;

                // if there was a carry...
                //
                if( result.low < low )

                    // believe it or not, the low word
                    // is already correct!!!
                    //

                    // fix the high dword
                    //
                    result.high++;

                return result;
            }

            // addition with assignment operator overload
            //
            Bit64 &operator +=( const Bit64 &second )
            {
                Bit32 low2 = low + second.low;
                high += second.high;

                // if there was a carry...
                //
                if( low2 < low )

                    // believe it or not, the low word
                    // is already correct!!!
                    //

                    // fix the high dword
                    //
                    high++;

                low = low2;
                return *this;
            }

#if 0
            // division operator overload
            //
            Bit64 operator /( const Bit64 &second ) const
            {
            }

            // division operator overload
            //
            Bit64 operator /( const Bit32 second ) const
            {
                Bit64 result;

                result.high = high / second.high;
???             result.low = DWORD_SIZE*(high % second.high) + second.low;

                return result;
            }
#endif

            // equality comparison operator overload
            //
            int operator ==( const Bit64 &second ) const
            {
                return low==second.low && high==second.high;
            }

            // less than or equal comparison operator overload
            //
            int operator <=( const Bit64 &second ) const
            {
                return high<second.high
                   || (high==second.high && low<=second.low);
            }

            // greater than or equal comparison operator overload
            //
            int operator >=( const Bit64 &second ) const
            {
                return high>second.high
                   || (high==second.high && low>=second.low);
            }

            // greater than comparison operator overload
            //
            int operator >( const Bit64 &second ) const
            {
                return high>second.high
                   || (high==second.high && low>second.low);
            }

            // less than comparison operator overload
            //
            int operator <( const Bit64 &second ) const
            {
                return high<second.high
                   || (high==second.high && low<second.low);
            }
        };
#    endif
#endif

typedef struct
{
    Bit32 offset;
    Bit16 selector;
} Bit48;

// pointer to a function that accepts
// no arguments and returns nothing
//
typedef void (*vfp)( void );

#endif      // not included yet
